import unittest

from pcs_test.tools.assertions import (
    AssertPcsMixin,
    ac,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import (
    PcsRunner,
    pcs,
)

# pylint: disable=invalid-name
# pylint: disable=no-self-use
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements

empty_cib = rc("cib-empty.xml")


class ACLTest(unittest.TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_acl")
        write_file_to_tmpfile(empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def testEnableDisable(self):
        o, r = pcs(self.temp_cib.name, "acl disable".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")

        o, r = pcs(self.temp_cib.name, "acl enable".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(o, "ACLs are enabled\n\n")

        o, r = pcs(self.temp_cib.name, "acl disable".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")

    def testUserGroupCreateDeleteWithRoles(self):
        o, r = pcs(
            self.temp_cib.name,
            "acl role create role1 read xpath /xpath1/ write xpath /xpath2/".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role2 deny xpath /xpath3/ deny xpath /xpath4/".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role3 read xpath /xpath5/ read xpath /xpath6/".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl user create user1 roleX".split())
        ac(o, "Error: ACL role 'roleX' does not exist\n")
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl user create user1 role1 roleX".split(),
        )
        ac(o, "Error: ACL role 'roleX' does not exist\n")
        self.assertEqual(1, r)

        o, r = pcs(self.temp_cib.name, "acl group create group1 roleX".split())
        ac(o, "Error: ACL role 'roleX' does not exist\n")
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl group create group1 role1 roleX".split(),
        )
        ac(o, "Error: ACL role 'roleX' does not exist\n")
        self.assertEqual(1, r)

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: read xpath /xpath6/ (role3-read-1)
""",
        )
        self.assertEqual(0, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl user create user1 role1 role2".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(
            self.temp_cib.name,
            "acl group create group1 role1 role3".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role1 role2
Group: group1
  Roles: role1 role3
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: read xpath /xpath6/ (role3-read-1)
""",
        )

        o, r = pcs(self.temp_cib.name, "acl role create group1".split())
        assert r == 1
        ac(o, "Error: 'group1' already exists\n")

        o, r = pcs(self.temp_cib.name, "acl role create role1".split())
        assert r == 1
        ac(o, "Error: 'role1' already exists\n")

        o, r = pcs(self.temp_cib.name, "acl user create user1".split())
        assert r == 1
        ac(o, "Error: 'user1' already exists\n")

        o, r = pcs(self.temp_cib.name, "acl group create group1".split())
        assert r == 1
        ac(o, "Error: 'group1' already exists\n")

        o, r = pcs(self.temp_cib.name, "acl group create role1".split())
        assert r == 1
        ac(o, "Error: 'role1' already exists\n")

        o, r = pcs(
            self.temp_cib.name,
            "acl role assign role1 to noexist".split(),
        )
        assert r == 1
        ac(o, "Error: ACL group/ACL user 'noexist' does not exist\n")

        o, r = pcs(
            self.temp_cib.name,
            "acl role assign noexist to user1".split(),
        )
        assert r == 1
        ac(o, "Error: ACL role 'noexist' does not exist\n")

        o, r = pcs(
            self.temp_cib.name,
            "acl role assign role3 to user1".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role1 role2 role3
Group: group1
  Roles: role1 role3
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: read xpath /xpath6/ (role3-read-1)
""",
        )

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign noexist from user1".split(),
        )
        assert r == 1
        ac(o, "Error: Role 'noexist' is not assigned to 'user1'\n")

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role3 from noexist".split(),
        )
        assert r == 1
        ac(o, "Error: ACL group/ACL user 'noexist' does not exist\n")

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role3 from user1".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role1 role2
Group: group1
  Roles: role1 role3
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: read xpath /xpath6/ (role3-read-1)
""",
        )

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role2 from user1".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role1 from user1".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
Group: group1
  Roles: role1 role3
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: read xpath /xpath6/ (role3-read-1)
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl role delete role3".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role assign role2 to user1".split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role2
Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )

        o, r = pcs(self.temp_cib.name, "acl role assign role1 user1".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role2 role1
Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role2 from user1 --autodelete".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role1
Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role unassign role1 from user1 --autodelete".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl user create user1 role1 role2".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role1 role2
Group: group1
  Roles: role1
Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role delete role1 --autodelete".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles: role2
Role: role2
  Permission: deny xpath /xpath3/ (role2-deny)
  Permission: deny xpath /xpath4/ (role2-deny-1)
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl role delete role2".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
""",
        )
        assert r == 0

    def testUserGroupCreateDelete(self):
        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")

        o, r = pcs(self.temp_cib.name, "acl user create user1".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl user create user2".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl user create user1".split())
        assert r == 1
        ac(o, "Error: 'user1' already exists\n")

        o, r = pcs(self.temp_cib.name, "acl group create group1".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl group create group2".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl group create group1".split())
        assert r == 1
        ac(o, "Error: 'group1' already exists\n")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
User: user2
  Roles:
Group: group1
  Roles:
Group: group2
  Roles:
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl group delete user1".split())
        assert r == 1
        ac(o, "Error: ACL group 'user1' does not exist\n")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
User: user2
  Roles:
Group: group1
  Roles:
Group: group2
  Roles:
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl group delete group2".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
User: user2
  Roles:
Group: group1
  Roles:
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl group remove group1".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user1
  Roles:
User: user2
  Roles:
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl user delete user1".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

User: user2
  Roles:
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl user remove user2".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        assert r == 0
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")

    def testRoleCreateDelete(self):
        o, r = pcs(self.temp_cib.name, "acl role create role0 read".split())
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 read //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 read xpath".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(self.temp_cib.name, "acl role create role0 read id".split())
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 readX xpath //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 read xpathX //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test read".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test read //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test read xpath".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test read id".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test readX xpath //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 description=test read xpathX //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test read".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test read //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test read xpath".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test read id".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test readX xpath //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role0 desc=test read xpathX //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl role create..."))
        self.assertEqual(1, r)

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")
        self.assertEqual(0, r)

        o, r = pcs(self.temp_cib.name, "acl role create role0".split())
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl role create role0".split())
        ac(o, "Error: 'role0' already exists\n")
        assert r == 1

        o, r = pcs(
            self.temp_cib.name,
            ["acl", "role", "create", "role0d", "description=empty role"],
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role1 read xpath /xpath/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            [
                "acl",
                "role",
                "create",
                "role2",
                "description=with description",
                "READ",
                "XPATH",
                "/xpath/",
            ],
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(
            self.temp_cib.name,
            (
                "acl role create role3 Read XPath /xpath_query/ wRiTe xpATH "
                "/xpath_query2/ deny xpath /xpath_query3/"
            ).split(),
        )
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role0
Role: role0d
  Description: empty role
Role: role1
  Permission: read xpath /xpath/ (role1-read)
Role: role2
  Description: with description
  Permission: read xpath /xpath/ (role2-read)
Role: role3
  Permission: read xpath /xpath_query/ (role3-read)
  Permission: write xpath /xpath_query2/ (role3-write)
  Permission: deny xpath /xpath_query3/ (role3-deny)
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl role delete role2".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role0
Role: role0d
  Description: empty role
Role: role1
  Permission: read xpath /xpath/ (role1-read)
Role: role3
  Permission: read xpath /xpath_query/ (role3-read)
  Permission: write xpath /xpath_query2/ (role3-write)
  Permission: deny xpath /xpath_query3/ (role3-deny)
""",
        )
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl role delete role2".split())
        assert r == 1
        ac(o, "Error: ACL role 'role2' does not exist\n")

        o, r = pcs(self.temp_cib.name, "acl role delete role1".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl role remove role3".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl role remove role0".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, "acl role remove role0d".split())
        assert r == 0
        ac(o, "")

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(o, "ACLs are disabled, run 'pcs acl enable' to enable\n\n")
        assert r == 0

    def testPermissionAddDelete(self):
        o, r = pcs(
            self.temp_cib.name,
            "acl role create role1 read xpath /xpath1/ write xpath /xpath2/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role2 read xpath /xpath3/ write xpath /xpath4/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl role create role3 read xpath /xpath5/ write xpath /xpath6/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl config".split())
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
Role: role2
  Permission: read xpath /xpath3/ (role2-read)
  Permission: write xpath /xpath4/ (role2-write)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: write xpath /xpath6/ (role3-write)
""",
        )

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 deny xpath /myxpath1/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role4 deny xpath /myxpath2/".split(),
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, "acl config".split())
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
  Permission: deny xpath /myxpath1/ (role1-deny)
Role: role2
  Permission: read xpath /xpath3/ (role2-read)
  Permission: write xpath /xpath4/ (role2-write)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: write xpath /xpath6/ (role3-write)
Role: role4
  Permission: deny xpath /myxpath2/ (role4-deny)
""",
        )

        o, r = pcs(
            self.temp_cib.name, "acl permission delete role4-deny".split()
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name, "acl permission delete role4-deny".split()
        )
        ac(o, "Error: ACL permission 'role4-deny' does not exist\n")
        assert r == 1

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role4-deny".split()
        )
        ac(o, "Error: ACL permission 'role4-deny' does not exist\n")
        assert r == 1

        o, r = pcs(self.temp_cib.name, "acl config".split())
        assert r == 0
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
  Permission: deny xpath /myxpath1/ (role1-deny)
Role: role2
  Permission: read xpath /xpath3/ (role2-read)
  Permission: write xpath /xpath4/ (role2-write)
Role: role3
  Permission: read xpath /xpath5/ (role3-read)
  Permission: write xpath /xpath6/ (role3-write)
Role: role4
""",
        )

        o, r = pcs(
            self.temp_cib.name, "acl permission delete role3-read".split()
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(
            self.temp_cib.name, "acl permission delete role3-write".split()
        )
        ac(o, "")
        assert r == 0

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
  Permission: read xpath /xpath1/ (role1-read)
  Permission: write xpath /xpath2/ (role1-write)
  Permission: deny xpath /myxpath1/ (role1-deny)
Role: role2
  Permission: read xpath /xpath3/ (role2-read)
  Permission: write xpath /xpath4/ (role2-write)
Role: role3
Role: role4
""",
        )
        assert r == 0

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role1-read".split()
        )
        ac(o, "")
        self.assertEqual(0, r)

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role1-write".split()
        )
        ac(o, "")
        self.assertEqual(0, r)

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role1-deny".split()
        )
        ac(o, "")
        self.assertEqual(0, r)

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role2-read".split()
        )
        ac(o, "")
        self.assertEqual(0, r)

        o, r = pcs(
            self.temp_cib.name, "acl permission remove role2-write".split()
        )
        ac(o, "")
        self.assertEqual(0, r)

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
Role: role2
Role: role3
Role: role4
""",
        )
        self.assertEqual(0, r)

        o, r = pcs(self.temp_cib.name, "acl permission add role1 read".split())
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read xpath".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read id".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 readX xpath //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read xpathX //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read id dummy read".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read id dummy read //resources".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read id dummy read xpath".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        o, r = pcs(
            self.temp_cib.name,
            "acl permission add role1 read id dummy read id".split(),
        )
        self.assertTrue(o.startswith("\nUsage: pcs acl permission add..."))
        self.assertEqual(1, r)

        self.assert_pcs_fail(
            "acl permission add role1 read id dummy readX xpath //resources".split(),
            stdout_start="\nUsage: pcs acl permission add...",
        )

        self.assert_pcs_fail(
            "acl permission add role1 read id dummy read xpathX //resources".split(),
            stdout_start="\nUsage: pcs acl permission add...",
        )

        o, r = pcs(self.temp_cib.name, ["acl"])
        ac(
            o,
            """\
ACLs are disabled, run 'pcs acl enable' to enable

Role: role1
Role: role2
Role: role3
Role: role4
""",
        )
        self.assertEqual(0, r)

    def test_can_add_permission_for_existing_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_success(
            "acl permission add role1 read id role2".split()
        )

    def test_can_add_permission_for_existing_xpath(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success(
            "acl permission add role1 read xpath //nodes".split()
        )

    def test_can_not_add_permission_for_nonexisting_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_fail(
            "acl permission add role1 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_add_permission_for_nonexisting_id_in_later_part(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_fail(
            "acl permission add role1 read id role2 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_add_permission_for_nonexisting_role_with_bad_id(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_fail(
            "acl permission add #bad-name read id role1".split(),
            "Error: invalid ACL role '#bad-name'"
            + ", '#' is not a valid first character for a ACL role\n",
        )

    def test_can_create_role_with_permission_for_existing_id(self):
        self.assert_pcs_success("acl role create role2".split())
        self.assert_pcs_success("acl role create role1 read id role2".split())

    def test_can_not_crate_role_with_permission_for_nonexisting_id(self):
        self.assert_pcs_fail(
            "acl role create role1 read id non-existent-id".split(),
            "Error: id 'non-existent-id' does not exist\n",
        )

    def test_can_not_create_role_with_bad_name(self):
        self.assert_pcs_fail(
            "acl role create #bad-name".split(),
            "Error: invalid ACL role '#bad-name'"
            + ", '#' is not a valid first character for a ACL role\n",
        )

    def test_fail_on_unknown_role_method(self):
        self.assert_pcs_fail(
            "acl role unknown whatever".split(),
            stdout_start="\nUsage: pcs acl role ...",
        )

    def test_assign_unassign_role_to_user(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_success("acl role assign role1 user user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 user user1".split(),
            "Error: Role 'role1' is already assigned to 'user1'\n",
        )
        self.assert_pcs_success("acl role unassign role1 user user1".split())
        self.assert_pcs_fail(
            "acl role unassign role1 user user1".split(),
            "Error: Role 'role1' is not assigned to 'user1'\n",
        )

    def test_assign_unassign_role_to_user_not_existing_user(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to user group1".split(),
            "Error: 'group1' is not an ACL user\n",
        )

    def test_assign_unassign_role_to_user_with_to(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_success("acl role assign role1 to user user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to user user1".split(),
            "Error: Role 'role1' is already assigned to 'user1'\n",
        )
        self.assert_pcs_success(
            "acl role unassign role1 from user user1".split()
        )
        self.assert_pcs_fail(
            "acl role unassign role1 from user user1".split(),
            "Error: Role 'role1' is not assigned to 'user1'\n",
        )

    def test_assign_unassign_role_to_group(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_success("acl role assign role1 group group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 group group1".split(),
            "Error: Role 'role1' is already assigned to 'group1'\n",
        )
        self.assert_pcs_success("acl role unassign role1 group group1".split())
        self.assert_pcs_fail(
            "acl role unassign role1 group group1".split(),
            "Error: Role 'role1' is not assigned to 'group1'\n",
        )

    def test_assign_unassign_role_to_group_not_existing_group(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl user create user1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to group user1".split(),
            "Error: ACL group 'user1' does not exist\n",
        )

    def test_assign_unassign_role_to_group_with_to(self):
        self.assert_pcs_success("acl role create role1".split())
        self.assert_pcs_success("acl group create group1".split())
        self.assert_pcs_success("acl role assign role1 to group group1".split())
        self.assert_pcs_fail(
            "acl role assign role1 to group group1".split(),
            "Error: Role 'role1' is already assigned to 'group1'\n",
        )
        self.assert_pcs_success(
            "acl role unassign role1 from group group1".split()
        )
        self.assert_pcs_fail(
            "acl role unassign role1 from group group1".split(),
            "Error: Role 'role1' is not assigned to 'group1'\n",
        )



"""
The intention is put there knowledge about cluster state structure.
Hide information about underlying xml is desired too.
"""
from collections import defaultdict
from typing import (
    Any,
    Dict,
)

from lxml import etree

from pcs.common import (
    const,
    reports,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.pacemaker.values import (
    is_false,
    is_true,
)
from pcs.lib.xml_tools import find_parent


class ResourceNotFound(Exception):
    pass


_id_xpath_predicate = "(@id=$id or starts-with(@id, concat($id, ':')))"


class _Attrs:
    def __init__(self, owner_name, attrib, required_attrs):
        """
        attrib lxml.etree._Attrib - wrapped attribute collection
        required_attrs dict of required attribute names object_name:xml_attribute
        """
        self.owner_name = owner_name
        self.attrib = attrib
        self.required_attrs = required_attrs

    def __getattr__(self, name):
        if name in self.required_attrs.keys():
            try:
                attr_specification = self.required_attrs[name]
                if isinstance(attr_specification, tuple):
                    attr_name, attr_transform = attr_specification
                    return attr_transform(self.attrib[attr_name])
                return self.attrib[attr_specification]
            except KeyError as e:
                raise AttributeError(
                    "Missing attribute '{0}' ('{1}' in source) in '{2}'".format(
                        name, self.required_attrs[name], self.owner_name
                    )
                ) from e

        raise AttributeError(
            "'{0}' does not declare attribute '{1}'".format(
                self.owner_name, name
            )
        )


class _Children:
    def __init__(self, owner_name, dom_part, children, sections):
        self.owner_name = owner_name
        self.dom_part = dom_part
        self.children = children
        self.sections = sections

    def __getattr__(self, name):
        if name in self.children.keys():
            element_name, wrapper = self.children[name]
            return [
                wrapper(element)
                for element in self.dom_part.xpath(
                    ".//*[local-name()=$tag_name]", tag_name=element_name
                )
            ]

        if name in self.sections.keys():
            element_name, wrapper = self.sections[name]
            return wrapper(
                self.dom_part.xpath(
                    ".//*[local-name()=$tag_name]", tag_name=element_name
                )[0]
            )

        raise AttributeError(
            "'{0}' does not declare child or section '{1}'".format(
                self.owner_name, name
            )
        )


class _Element:
    # Note: not properly typed
    required_attrs: Dict[Any, Any] = {}
    # Note: not properly typed
    children: Dict[Any, Any] = {}
    # Note: not properly typed
    sections: Dict[Any, Any] = {}

    def __init__(self, dom_part: etree._Element):
        self.dom_part = dom_part
        self.attrs = _Attrs(
            self.__class__.__name__, self.dom_part.attrib, self.required_attrs
        )
        self.children_access = _Children(
            self.__class__.__name__,
            self.dom_part,
            self.children,
            self.sections,
        )

    def __getattr__(self, name):
        return getattr(self.children_access, name)


class _SummaryNodes(_Element):
    required_attrs = {
        "count": ("number", int),
    }


class _SummaryResources(_Element):
    required_attrs = {
        "count": ("number", int),
    }


class _SummarySection(_Element):
    sections = {
        "nodes": ("nodes_configured", _SummaryNodes),
        "resources": ("resources_configured", _SummaryResources),
    }


class _Node(_Element):
    required_attrs = {
        "id": "id",
        "name": "name",
        "type": "type",
        "online": ("online", is_true),
        "standby": ("standby", is_true),
        "standby_onfail": ("standby_onfail", is_true),
        "maintenance": ("maintenance", is_true),
        "pending": ("pending", is_true),
        "unclean": ("unclean", is_true),
        "shutdown": ("shutdown", is_true),
        "expected_up": ("expected_up", is_true),
        "is_dc": ("is_dc", is_true),
        "resources_running": ("resources_running", int),
    }


class _NodeSection(_Element):
    children = {
        "nodes": ("node", _Node),
    }


class ClusterState(_Element):
    sections = {
        "summary": ("summary", _SummarySection),
        "node_section": ("nodes", _NodeSection),
    }


def _get_primitives_for_state_check(
    cluster_state, resource_id, expected_running
):
    primitives = cluster_state.xpath(
        """
        .//resource[{predicate_id}]
        |
        .//group[{predicate_id}]/resource[{predicate_position}]
        |
        .//clone[@id=$id]/resource
        |
        .//clone[@id=$id]/group/resource[{predicate_position}]
        |
        .//bundle[@id=$id]/replica/resource
    """.format(
            predicate_id=_id_xpath_predicate,
            predicate_position=("last()" if expected_running else "1"),
        ),
        id=resource_id,
    )
    return [
        element
        for element in primitives
        if not is_true(element.attrib.get("failed", ""))
    ]


def _get_primitive_roles_with_nodes(primitive_el_list):
    # Clone resources are represented by multiple primitive elements.
    roles_with_nodes = defaultdict(set)
    for resource_element in primitive_el_list:
        if resource_element.attrib["role"] in const.PCMK_ROLES_RUNNING:
            roles_with_nodes[resource_element.attrib["role"]].update(
                [
                    node.attrib["name"]
                    for node in resource_element.findall(".//node")
                ]
            )
    return {role: sorted(nodes) for role, nodes in roles_with_nodes.items()}


def get_resource_state(cluster_state, resource_id):
    return _get_primitive_roles_with_nodes(
        _get_primitives_for_state_check(
            cluster_state, resource_id, expected_running=True
        )
    )


def info_resource_state(cluster_state, resource_id):
    roles_with_nodes = get_resource_state(cluster_state, resource_id)
    if not roles_with_nodes:
        return ReportItem.info(reports.messages.ResourceDoesNotRun(resource_id))
    return ReportItem.info(
        reports.messages.ResourceRunningOnNodes(resource_id, roles_with_nodes)
    )


def ensure_resource_state(expected_running, cluster_state, resource_id):
    roles_with_nodes = _get_primitive_roles_with_nodes(
        _get_primitives_for_state_check(
            cluster_state, resource_id, expected_running
        )
    )
    if not roles_with_nodes:
        return ReportItem(
            reports.item.ReportItemSeverity(
                reports.ReportItemSeverity.INFO
                if not expected_running
                else reports.ReportItemSeverity.ERROR
            ),
            reports.messages.ResourceDoesNotRun(resource_id),
        )
    return ReportItem(
        reports.item.ReportItemSeverity(
            reports.ReportItemSeverity.INFO
            if expected_running
            else reports.ReportItemSeverity.ERROR
        ),
        reports.messages.ResourceRunningOnNodes(resource_id, roles_with_nodes),
    )


def ensure_resource_running(cluster_state, resource_id):
    return ensure_resource_state(
        expected_running=True,
        cluster_state=cluster_state,
        resource_id=resource_id,
    )


def is_resource_managed(cluster_state, resource_id):
    """
    Check if the resource is managed

    etree cluster_state -- status of the cluster
    string resource_id -- id of the resource
    """
    primitive_list = cluster_state.xpath(
        """
        .//resource[{predicate_id}]
        |
        .//group[{predicate_id}]/resource
        """.format(
            predicate_id=_id_xpath_predicate
        ),
        id=resource_id,
    )
    if primitive_list:
        for primitive in primitive_list:
            if is_false(primitive.attrib.get("managed", "")):
                return False
            parent = find_parent(primitive, ["clone", "bundle"])
            if parent is not None and is_false(
                parent.attrib.get("managed", "")
            ):
                return False
        return True

    parent_list = cluster_state.xpath(
        """
        .//clone[@id=$resource_id]
        |
        .//bundle[@id=$resource_id]
        """,
        resource_id=resource_id,
    )
    for parent in parent_list:
        if is_false(parent.attrib.get("managed", "")):
            return False
        for primitive in parent.xpath(".//resource"):
            if is_false(primitive.attrib.get("managed", "")):
                return False
        return True

    raise ResourceNotFound(resource_id)



from .types import (
    AddRemoveContainerType,
    AddRemoveItemType,
    BoothConfigUsedWhere,
    DefaultAddressSource,
    FenceHistoryCommandType,
    PcsCommand,
    ReasonType,
    ServiceAction,
    StonithRestartlessUpdateUnableToPerformReason,
    StonithWatchdogTimeoutCannotBeSetReason,
)

ADD_REMOVE_CONTAINER_TYPE_STONITH_RESOURCE = AddRemoveContainerType("stonith")
ADD_REMOVE_CONTAINER_TYPE_GROUP = AddRemoveContainerType("group")
ADD_REMOVE_ITEM_TYPE_DEVICE = AddRemoveItemType("device")
ADD_REMOVE_ITEM_TYPE_RESOURCE = AddRemoveItemType("resource")
BOOTH_CONFIG_USED_IN_CLUSTER_RESOURCE = BoothConfigUsedWhere(
    "in a cluster resource"
)
BOOTH_CONFIG_USED_ENABLED_IN_SYSTEMD = BoothConfigUsedWhere(
    "enabled in systemd"
)
BOOTH_CONFIG_USED_RUNNING_IN_SYSTEMD = BoothConfigUsedWhere(
    "running in systemd"
)

FENCE_HISTORY_COMMAND_CLEANUP = FenceHistoryCommandType("cleanup")
FENCE_HISTORY_COMMAND_SHOW = FenceHistoryCommandType("show")
FENCE_HISTORY_COMMAND_UPDATE = FenceHistoryCommandType("update")

PCS_COMMAND_OPERATION_DEFAULTS_UPDATE = PcsCommand(
    "resource op defaults update"
)
PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE = PcsCommand("resource defaults update")

SERVICE_ACTION_START = ServiceAction("START")
SERVICE_ACTION_STOP = ServiceAction("STOP")
SERVICE_ACTION_ENABLE = ServiceAction("ENABLE")
SERVICE_ACTION_DISABLE = ServiceAction("DISABLE")
SERVICE_ACTION_KILL = ServiceAction("KILL")

REASON_UNREACHABLE = ReasonType("unreachable")
REASON_NOT_LIVE_CIB = ReasonType("not_live_cib")

DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS = DefaultAddressSource("known_hosts")
DEFAULT_ADDRESS_SOURCE_HOST_NAME = DefaultAddressSource("host_name")

SBD_NOT_SET_UP = StonithWatchdogTimeoutCannotBeSetReason("sbd_not_set_up")
SBD_SET_UP_WITH_DEVICES = StonithWatchdogTimeoutCannotBeSetReason(
    "sbd_set_up_with_devices"
)
SBD_SET_UP_WITHOUT_DEVICES = StonithWatchdogTimeoutCannotBeSetReason(
    "sbd_set_up_without_devices"
)

STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_NOT_RUNNING = (
    StonithRestartlessUpdateUnableToPerformReason("not_running")
)
STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_OTHER = (
    StonithRestartlessUpdateUnableToPerformReason("other")
)



from pcs.common import file_type_codes
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.toolbox import FileToolbox
from pcs.lib.file.toolbox import for_file_type as get_file_toolbox

from .config.facade import (
    DrRole,
    Facade,
)


class DrEnv:
    def __init__(self) -> None:
        self._config_file = FileInstance.for_dr_config()

    @staticmethod
    def create_facade(role: DrRole) -> Facade:
        return Facade.create(role)

    @property
    def config(self) -> FileInstance:
        return self._config_file

    @staticmethod
    def get_config_toolbox() -> FileToolbox:
        return get_file_toolbox(file_type_codes.PCS_DR_CONFIG)



from typing import (
    Optional,
    Sequence,
)

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto


@dataclass(frozen=True)
class CibResourceCloneDto(DataTransferObject):
    id: str  # pylint: disable=invalid-name
    description: Optional[str]
    member_id: str
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs import settings
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib.cib import node
from pcs.lib.cib.tools import IdProvider
from pcs.lib.pacemaker.state import ClusterState

from pcs_test.tools import fixture_crm_mon
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str

# pylint: disable=protected-access


@mock.patch("pcs.lib.cib.node._ensure_node_exists")
class UpdateNodeInstanceAttrs(TestCase):
    def setUp(self):
        self.node1 = etree.fromstring(
            """
            <node id="1" uname="rh73-node1"/>
        """
        )
        self.node2 = etree.fromstring(
            """
            <node id="2" uname="rh73-node2">
                <instance_attributes id="nodes-2">
                    <nvpair name="a" value="A" />
                    <nvpair name="b" value="B" />
                    <nvpair name="c" value="C" />
                </instance_attributes>
            </node>
        """
        )
        self.node3 = etree.fromstring(
            """
            <node id="3" uname="rh73-node3" >
                <instance_attributes id="nodes-3-a">
                    <nvpair name="a" value="A" />
                </instance_attributes>
                <instance_attributes id="nodes-3-b">
                    <nvpair name="b" value="B" />
                </instance_attributes>
            </node>
        """
        )
        self.cib = etree.fromstring(
            self.compile_cib(self.node1, self.node2, self.node3)
        )
        self.state = "node state list"
        self.id_provider = IdProvider(self.cib)

    @staticmethod
    def compile_cib(*node_list):
        parts = ["<cib><configuration><nodes>"]
        parts += [
            node if isinstance(node, str) else etree_to_str(node)
            for node in node_list
        ]
        parts.append("</nodes></configuration></cib>")
        return "".join(parts)

    def test_empty_node(self, mock_get_node):
        mock_get_node.return_value = self.node1
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node1",
            {"x": "X"},
            state_nodes=self.state,
        )
        assert_xml_equal(
            etree_to_str(self.node1),
            """
                <node id="1" uname="rh73-node1">
                    <instance_attributes id="nodes-1">
                        <nvpair id="nodes-1-x" name="x" value="X" />
                    </instance_attributes>
                </node>
            """,
        )

    def test_existing_attrs(self, mock_get_node):
        mock_get_node.return_value = self.node2
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node2",
            {"a": "", "b": "b", "x": "X"},
            state_nodes=self.state,
        )
        assert_xml_equal(
            etree_to_str(self.node2),
            """
                <node id="2" uname="rh73-node2">
                    <instance_attributes id="nodes-2">
                        <nvpair name="b" value="b" />
                        <nvpair name="c" value="C" />
                        <nvpair id="nodes-2-x" name="x" value="X" />
                    </instance_attributes>
                </node>
            """,
        )

    def test_multiple_attrs_sets(self, mock_get_node):
        mock_get_node.return_value = self.node3
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node3",
            {"x": "X"},
            state_nodes=self.state,
        )
        assert_xml_equal(
            etree_to_str(self.node3),
            """
                <node id="3" uname="rh73-node3" >
                    <instance_attributes id="nodes-3-a">
                        <nvpair name="a" value="A" />
                        <nvpair id="nodes-3-a-x" name="x" value="X" />
                    </instance_attributes>
                    <instance_attributes id="nodes-3-b">
                        <nvpair name="b" value="B" />
                    </instance_attributes>
                </node>
            """,
        )

    def test_keep_empty_nvset_and_node(self, mock_get_node):
        mock_get_node.return_value = self.node2
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node2",
            {"a": "", "b": "", "c": ""},
            state_nodes=self.state,
        )
        assert_xml_equal(
            self.compile_cib(
                self.node1,
                """
                    <node id="2" uname="rh73-node2">
                        <instance_attributes id="nodes-2" />
                    </node>
                """,
                self.node3,
            ),
            self.compile_cib(self.node1, self.node2, self.node3),
        )

    def test_dont_create_empty_nvset_if_deleting(self, mock_get_node):
        mock_get_node.return_value = self.node1
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node1",
            {"x": ""},
            state_nodes=self.state,
        )
        assert_xml_equal(
            etree_to_str(self.node1), """<node id="1" uname="rh73-node1" />"""
        )

    def test_dont_create_empty_nvset_if_no_attrs(self, mock_get_node):
        mock_get_node.return_value = self.node1
        node.update_node_instance_attrs(
            self.cib, self.id_provider, "rh73-node1", {}, state_nodes=self.state
        )
        assert_xml_equal(
            etree_to_str(self.node1), """<node id="1" uname="rh73-node1" />"""
        )

    def test_dont_create_empty_node_if_deleting(self, mock_get_node):
        node4 = etree.fromstring("""<node id="4" />""")
        mock_get_node.return_value = node4
        node.update_node_instance_attrs(
            self.cib,
            self.id_provider,
            "rh73-node4",
            {"x": ""},
            state_nodes=self.state,
        )
        assert_xml_equal(
            etree_to_str(self.cib),
            self.compile_cib(self.node1, self.node2, self.node3),
        )


class EnsureNodeExists(TestCase):
    def setUp(self):
        self.node1 = etree.fromstring(
            """
            <node id="1" uname="name-test1" type="member" />
        """
        )
        self.node2 = etree.fromstring(
            """
            <node id="2" uname="name-test2" type="member" />
        """
        )
        self.nodes = etree.Element("nodes")
        self.nodes.append(self.node1)

    @staticmethod
    def fixture_state():
        with open(rc("crm_mon.minimal.xml")) as crm_mon_file:
            crm_mon_xml = crm_mon_file.read()
        return ClusterState(
            fixture_crm_mon.complete_state(
                crm_mon_xml,
                nodes_xml="""
                <nodes>
                    <node name="name-test1" id="1" is_dc="true" />
                    <node name="name-test2" id="2" />
                </nodes>
            """,
            )
        ).node_section.nodes

    def test_node_already_exists(self):
        assert_xml_equal(
            etree_to_str(node._ensure_node_exists(self.nodes, "name-test1")),
            etree_to_str(self.node1),
        )

    def test_node_missing_no_state(self):
        assert_raise_library_error(
            lambda: node._ensure_node_exists(self.nodes, "name-missing"),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {"node": "name-missing", "searched_types": []},
                None,
            ),
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_node_missing_not_in_state(self):
        assert_raise_library_error(
            lambda: node._ensure_node_exists(
                self.nodes, "name-missing", self.fixture_state()
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {"node": "name-missing", "searched_types": []},
                None,
            ),
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_node_missing_and_gets_created(self):
        assert_xml_equal(
            etree_to_str(
                node._ensure_node_exists(
                    self.nodes, "name-test2", self.fixture_state()
                )
            ),
            etree_to_str(self.node2),
        )


class GetNodeByUname(TestCase):
    def setUp(self):
        self.node1 = etree.fromstring(
            """
            <node id="id-test1" uname="name-test1" />
        """
        )
        self.node2 = etree.fromstring(
            """
            <node id="id-test2" uname="name-test2" />
        """
        )
        self.nodes = etree.Element("nodes")
        self.nodes.append(self.node1)
        self.nodes.append(self.node2)

    def test_found(self):
        assert_xml_equal(
            etree_to_str(node._get_node_by_uname(self.nodes, "name-test1")),
            """<node id="id-test1" uname="name-test1" />""",
        )

    def test_not_found(self):
        self.assertTrue(node._get_node_by_uname(self.nodes, "id-test1") is None)


class CreateNode(TestCase):
    def setUp(self):
        self.nodes = etree.Element("nodes")

    def test_minimal(self):
        new_node = node._create_node("id-test", "name-test")
        self.nodes.append(new_node)
        assert_xml_equal(
            """
            <nodes>
                <node id="id-test" uname="name-test" />
            </nodes>
            """,
            etree_to_str(self.nodes),
        )

    def test_with_type(self):
        new_node = node._create_node("id-test", "name-test", "type-test")
        self.nodes.append(new_node)
        assert_xml_equal(
            """
            <nodes>
                <node id="id-test" uname="name-test" type="type-test" />
            </nodes>
            """,
            etree_to_str(self.nodes),
        )



from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import prepare_options
from pcs.cli.resource.parse_args import (
    parse_create_simple as parse_resource_create_args,
)


def _node_add_remote_separate_name_and_addr(arg_list):
    """
    Commandline options: no options
    """
    node_name = arg_list[0]
    if len(arg_list) == 1:
        node_addr = None
        rest_args = []
    elif "=" in arg_list[1] or arg_list[1] in ["op", "meta"]:
        node_addr = None
        rest_args = arg_list[1:]
    else:
        node_addr = arg_list[1]
        rest_args = arg_list[2:]
    return node_name, node_addr, rest_args


def node_add_remote(lib, arg_list, modifiers):
    """
    Options:
      * --wait
      * --force - allow incomplete distribution of files, allow pcmk remote
        service to fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      * --no-default-ops - do not use default operations
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
        "--no-default-ops",
    )
    if not arg_list:
        raise CmdLineInputError()

    node_name, node_addr, rest_args = _node_add_remote_separate_name_and_addr(
        arg_list
    )

    parts = parse_resource_create_args(rest_args)
    force = modifiers.get("--force")

    lib.remote_node.node_add_remote(
        node_name,
        node_addr,
        parts["op"],
        parts["meta"],
        parts["options"],
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_incomplete_distribution=force,
        allow_pacemaker_remote_service_fail=force,
        allow_invalid_operation=force,
        allow_invalid_instance_attributes=force,
        use_default_operations=not modifiers.get("--no-default-ops"),
        wait=modifiers.get("--wait"),
    )


def create_node_remove_remote(remove_resource):
    def node_remove_remote(lib, arg_list, modifiers):
        """
        Options:
          * --force - allow multiple nodes removal, allow pcmk remote service
            to fail, don't stop a resource before its deletion (this is side
            effect of old resource delete command used here)
          * --skip-offline - skip offline nodes
          * --request-timeout - HTTP request timeout
          For tests:
          * --corosync_conf
          * -f
        """
        modifiers.ensure_only_supported(
            "--force",
            "--skip-offline",
            "--request-timeout",
            "--corosync_conf",
            "-f",
        )
        if len(arg_list) != 1:
            raise CmdLineInputError()
        lib.remote_node.node_remove_remote(
            arg_list[0],
            remove_resource,
            skip_offline_nodes=modifiers.get("--skip-offline"),
            allow_remove_multiple_nodes=modifiers.get("--force"),
            allow_pacemaker_remote_service_fail=modifiers.get("--force"),
        )

    return node_remove_remote


def node_add_guest(lib, arg_list, modifiers):
    """
    Options:
      * --wait
      * --force - allow incomplete distribution of files, allow pcmk remote
        service to fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
    )
    if len(arg_list) < 2:
        raise CmdLineInputError()

    node_name = arg_list[0]
    resource_id = arg_list[1]
    meta_options = prepare_options(arg_list[2:])

    lib.remote_node.node_add_guest(
        node_name,
        resource_id,
        meta_options,
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_incomplete_distribution=modifiers.get("--force"),
        allow_pacemaker_remote_service_fail=modifiers.get("--force"),
        wait=modifiers.get("--wait"),
    )


def node_remove_guest(lib, arg_list, modifiers):
    """
    Options:
      * --wait
      * --force - allow multiple nodes removal, allow pcmk remote service to
        fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
    )
    if len(arg_list) != 1:
        raise CmdLineInputError()

    lib.remote_node.node_remove_guest(
        arg_list[0],
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_remove_multiple_nodes=modifiers.get("--force"),
        allow_pacemaker_remote_service_fail=modifiers.get("--force"),
        wait=modifiers.get("--wait"),
    )


def node_clear(lib, arg_list, modifiers):
    """
    Options:
      * --force - allow to clear a cluster node
    """
    modifiers.ensure_only_supported("--force")
    if len(arg_list) != 1:
        raise CmdLineInputError()

    lib.cluster.node_clear(
        arg_list[0], allow_clear_cluster_node=modifiers.get("--force")
    )



import json
from typing import (
    Dict,
    Iterable,
)

from dacite import DaciteError

from pcs.common import reports
from pcs.common.communication import const
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.interface.dto import from_dict
from pcs.common.node_communicator import (
    Request,
    RequestData,
)
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    MarkSuccessfulMixin,
    RunRemotelyBase,
)
from pcs.lib.node_communication import response_to_report_item


class Unfence(
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    MarkSuccessfulMixin,
    RunRemotelyBase,
):
    def __init__(
        self,
        report_processor: reports.ReportProcessor,
        original_devices: Iterable[str],
        updated_devices: Iterable[str],
    ) -> None:
        super().__init__(report_processor)
        self._original_devices = original_devices
        self._updated_devices = updated_devices

    def _get_request_data(self):
        return None

    def _prepare_initial_requests(self):
        return [
            Request(
                target,
                RequestData(
                    "api/v1/scsi-unfence-node/v2",
                    data=json.dumps(
                        dict(
                            node=target.label,
                            original_devices=self._original_devices,
                            updated_devices=self._updated_devices,
                        )
                    ),
                ),
            )
            for target in self._target_list
        ]

    def _process_response(self, response):
        report_item = response_to_report_item(
            response, report_pcsd_too_old_on_404=True
        )
        if report_item:
            self._report(report_item)
            return
        node_label = response.request.target.label
        try:
            result = from_dict(
                InternalCommunicationResultDto, json.loads(response.data)
            )
            context = reports.ReportItemContext(node_label)
            self._report_list(
                [
                    reports.report_dto_to_item(report, context)
                    for report in result.report_list
                ]
            )
            if result.status == const.COM_STATUS_SUCCESS:
                self._on_success()

        except (json.JSONDecodeError, DaciteError):
            self._report(
                reports.ReportItem.error(
                    reports.messages.InvalidResponseFormat(node_label)
                )
            )


class UnfenceMpath(Unfence):
    def __init__(
        self,
        report_processor: reports.ReportProcessor,
        original_devices: Iterable[str],
        updated_devices: Iterable[str],
        node_key_map: Dict[str, str],
    ) -> None:
        super().__init__(report_processor, original_devices, updated_devices)
        self._node_key_map = node_key_map

    def _prepare_initial_requests(self):
        return [
            Request(
                target,
                RequestData(
                    "api/v1/scsi-unfence-node-mpath/v1",
                    data=json.dumps(
                        dict(
                            key=self._node_key_map[target.label],
                            original_devices=self._original_devices,
                            updated_devices=self._updated_devices,
                        )
                    ),
                ),
            )
            for target in self._target_list
        ]



from unittest import TestCase

from pcs.common.reports import codes as report_codes
from pcs.lib.cib.node import PacemakerNode as PNode
from pcs.lib.corosync import config_validators
from pcs.lib.corosync.node import CorosyncNode as CNode
from pcs.lib.corosync.node import CorosyncNodeAddress as CAddr

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.custom_mock import patch_getaddrinfo

# pylint: disable=no-self-use

forbidden_characters_kwargs = dict(
    allowed_values=None,
    cannot_be_empty=False,
    forbidden_characters=r"{}\n\r",
)


class AddNodes(TestCase):
    # pylint: disable=too-many-public-methods
    fixture_coronodes_1_link = [
        CNode("node1", [CAddr("addr01", 1)], 1),
        CNode("node2", [CAddr("addr02", 1)], 2),
    ]

    fixture_coronodes_2_links = [
        CNode("node1", [CAddr("addr01", 1), CAddr("addr11", 2)], 1),
        CNode("node2", [CAddr("addr02", 1), CAddr("addr12", 2)], 2),
    ]

    def setUp(self):
        self.known_addrs = patch_getaddrinfo(
            self,
            [f"addr{i:02d}" for i in range(1, 20)]
            + [f"10.0.0.{i}" for i in range(1, 20)]
            + [f"::ffff:10:0:0:{i}" for i in range(1, 20)],
        )

    def test_all_valid_one_node_one_link(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [],
        )

    def test_all_more_nodes_more_links(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03", "addr13"]},
                    {"name": "node4", "addrs": ["addr04", "addr14"]},
                ],
                self.fixture_coronodes_2_links,
                [PNode("node-remote", "addr19")],
            ),
            [],
        )

    def test_nodelist_empty(self):
        assert_report_item_list_equal(
            config_validators.add_nodes([], self.fixture_coronodes_1_link, []),
            [fixture.error(report_codes.COROSYNC_NODES_MISSING)],
        )

    def test_empty_node(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                    {},
                    {"name": "node4", "addrs": ["addr04"]},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["name"],
                    option_type="node 2",
                ),
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=0,
                    min_count=1,
                    max_count=1,
                    node_name=None,
                    node_index=2,
                ),
            ],
        )

    def test_node_options_invalid(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"], "nonsense": "abc"},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["nonsense"],
                    option_type="node",
                    allowed=["addrs", "name"],
                    allowed_patterns=[],
                ),
            ],
        )

    def test_nodename_invalid(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "", "addrs": ["addr03"]},
                    {"addrs": ["addr04"]},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="",
                    option_name="node 1 name",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["name"],
                    option_type="node 2",
                ),
            ],
        )

    def test_nodename_not_unique(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                    {"name": "node3", "addrs": ["addr04"]},
                    # invalid nodes are not reported as duplicate
                    {"name": "", "addrs": ["addr05"]},
                    {"name": "", "addrs": ["addr06"]},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="",
                    option_name="node 3 name",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="",
                    option_name="node 4 name",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.NODE_NAMES_DUPLICATION, name_list=["node3"]
                ),
            ],
        )

    def test_nodename_already_used(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                    {"name": "node2", "addrs": ["addr04"]},
                    {"name": "node-remote", "addrs": ["addr05"]},
                ],
                [
                    CNode("node1", [CAddr("addr01", 1)], 1),
                    CNode("node2", [CAddr("addr02", 1)], 2),
                ],
                [PNode("node-remote", "addr19")],
            ),
            [
                fixture.error(
                    report_codes.NODE_NAMES_ALREADY_EXIST,
                    name_list=["node-remote", "node2"],
                )
            ],
        )

    def test_node_addrs_missing(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3"},
                    {"name": "node4", "addrs": []},
                    {"name": "node5", "addrs": None},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=0,
                    min_count=1,
                    max_count=1,
                    node_name=name,
                    node_index=id,
                )
                for id, name in enumerate(["node3", "node4", "node5"], 1)
            ],
        )

    def test_node_addrs_count_mismatch(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                    {"name": "node4", "addrs": ["addr04", "addr14"]},
                    {"name": "node5", "addrs": ["addr05", "addr15", "addr16"]},
                ],
                self.fixture_coronodes_2_links,
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=1,
                    min_count=2,
                    max_count=2,
                    node_name="node3",
                    node_index=1,
                ),
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=3,
                    min_count=2,
                    max_count=2,
                    node_name="node5",
                    node_index=3,
                ),
            ],
        )

    def test_node_addr_empty(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["", "addr13"]},
                    {"name": "node4", "addrs": ["addr04", "addr14"]},
                    {"name": "node5", "addrs": ["addr05", ""]},
                    {"name": "node6", "addrs": ["", ""]},
                    {"name": None, "addrs": ["", ""]},
                ],
                self.fixture_coronodes_2_links,
                [],
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_CANNOT_BE_EMPTY,
                    node_name_list=["node3", "node5", "node6"],
                ),
            ],
        )

    def test_node_addrs_unresolvable(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    # Duplicated addresses reported only once but they trigger
                    # a duplicate addresses report.
                    {"name": "node3", "addrs": ["addr03", "addrX2"]},
                    {"name": "node4", "addrs": ["addrX2", "addr14"]},
                    # Extra address reported as well, it triggres its own report
                    # about being an extra address.
                    {"name": "node5", "addrs": ["addr05", "addrX1", "addrX3"]},
                ],
                self.fixture_coronodes_2_links,
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=3,
                    min_count=2,
                    max_count=2,
                    node_name="node5",
                    node_index=3,
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=["addrX1", "addrX2", "addrX3"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_DUPLICATION,
                    address_list=["addrX2"],
                ),
            ],
        )

    def test_node_addrs_unresolvable_forced(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    # Duplicated addresses reported only once but they trigger
                    # a duplicate addresses report.
                    {"name": "node3", "addrs": ["addr03", "addrX2"]},
                    {"name": "node4", "addrs": ["addrX2", "addr14"]},
                    # Extra address reported as well, it triggres its own report
                    # about being an extra address.
                    {"name": "node5", "addrs": ["addr05", "addrX1", "addrX3"]},
                ],
                self.fixture_coronodes_2_links,
                [],
                force_unresolvable=True,
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=3,
                    min_count=2,
                    max_count=2,
                    node_name="node5",
                    node_index=3,
                ),
                fixture.warn(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=["addrX1", "addrX2", "addrX3"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_DUPLICATION,
                    address_list=["addrX2"],
                ),
            ],
        )

    def test_node_addrs_not_unique(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {
                        "name": "node3",
                        "addrs": ["addr03", "10.0.0.3", "::ffff:10:0:0:3"],
                    },
                    {
                        "name": "node4",
                        "addrs": ["addr04", "10.0.0.4", "::ffff:10:0:0:4"],
                    },
                    {
                        "name": "node5",
                        "addrs": ["addr04", "10.0.0.3", "::ffff:10:0:0:6"],
                    },
                    {
                        "name": "node6",
                        "addrs": ["addr06", "10.0.0.3", "::ffff:10:0:0:6"],
                    },
                ],
                [
                    CNode(
                        "node1",
                        [
                            CAddr("addr01", 1),
                            CAddr("10.0.0.1", 2),
                            CAddr("::ffff:10:0:0:1", 3),
                        ],
                        1,
                    ),
                    CNode(
                        "node2",
                        [
                            CAddr("addr02", 1),
                            CAddr("10.0.0.2", 2),
                            CAddr("::ffff:10:0:0:2", 3),
                        ],
                        2,
                    ),
                ],
                [],
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_DUPLICATION,
                    address_list=["10.0.0.3", "::ffff:10:0:0:6", "addr04"],
                )
            ],
        )

    def test_node_addrs_already_used(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["addr03"]},
                    {"name": "node4", "addrs": ["addr02"]},
                    {"name": "node5", "addrs": ["addr19"]},
                ],
                [
                    CNode("node1", [CAddr("addr01", 1)], 1),
                    CNode("node2", [CAddr("addr02", 1)], 2),
                ],
                [PNode("node-remote", "addr19")],
            ),
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["addr02", "addr19"],
                )
            ],
        )

    def test_node_addrs_ip_version_ok(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {
                        "name": "node3",
                        "addrs": [
                            "addr03",
                            "::ffff:10:0:0:3",
                            "10.0.0.3",
                            "addr13",
                        ],
                    },
                    {
                        "name": "node4",
                        "addrs": [
                            "10.0.0.4",
                            "addr04",
                            "addr14",
                            "::ffff:10:0:0:4",
                        ],
                    },
                ],
                [
                    CNode(
                        "node1",
                        [
                            CAddr("addr01", 1),
                            CAddr("addr11", 2),
                            CAddr("10.0.0.1", 5),
                            CAddr("::ffff:10:0:0:1", 6),
                        ],
                        1,
                    ),
                    CNode(
                        "node2",
                        [
                            CAddr("addr02", 1),
                            CAddr("addr12", 2),
                            CAddr("10.0.0.2", 5),
                            CAddr("::ffff:10:0:0:2", 6),
                        ],
                        2,
                    ),
                ],
                [],
            ),
            [],
        )

    def test_node_addrs_ip_version_mismatch(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["::ffff:10:0:0:3"]},
                    {"name": "node4", "addrs": ["10.0.0.14"]},
                ],
                [
                    CNode("node1", [CAddr("addr01", 1)], 1),
                    CNode("node2", [CAddr("addr02", 1)], 2),
                ],
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[1],
                )
            ],
        )

    def test_node_addrs_mismatch_existing_links(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node3", "addrs": ["::ffff:10:0:0:3", "addr13"]},
                    {"name": "node4", "addrs": ["addr04", "10.0.0.14"]},
                ],
                [
                    CNode(
                        "node1", [CAddr("10.0.0.1", 1), CAddr("addr11", 2)], 1
                    ),
                    CNode(
                        "node2",
                        [CAddr("addr02", 1), CAddr("::ffff:10:0:0:2", 2)],
                        2,
                    ),
                ],
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address="::ffff:10:0:0:3",
                    expected_address_type="IPv4",
                    link_number=1,
                ),
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address="10.0.0.14",
                    expected_address_type="IPv6",
                    link_number=2,
                ),
            ],
        )

    def test_node_addrs_ip_version_mismatch_complex(self):
        # several cases tested:
        # * not all links are defined - testing link indexes in reports
        # * link 1 - unresolvable addresses do not trigger ip mismatch reports
        # * link 3 - mismatch between new nodes only
        # * link 5 - mismatch between a new node and existing nodes,
        #   existing nodes mix FQDNs and IPs
        # * link 7 - mismatch between new nodes and new and existing node, only
        #   one report produced, existing nodes mix FQDNs and IPs
        # * node1 has an extra address which dosn't cause a crash trying to
        #   match it to a nonexisting link
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {
                        "name": "node3",
                        "addrs": [
                            "addrX1",
                            "::ffff:10:0:0:13",
                            "::ffff:10:0:0:3",
                            "10.0.0.3",
                            "10.0.0.13",
                        ],
                    },
                    {
                        "name": "node4",
                        "addrs": [
                            "addr04",
                            "10.0.0.4",
                            "addr14",
                            "::ffff:10:0:0:4",
                        ],
                    },
                ],
                [
                    CNode(
                        "node1",
                        [
                            CAddr("10.0.0.1", 1),
                            CAddr("addr11", 3),
                            CAddr("addr19", 5),
                            CAddr("addr18", 7),
                        ],
                        1,
                    ),
                    CNode(
                        "node2",
                        [
                            CAddr("addr02", 1),
                            CAddr("addr12", 3),
                            CAddr("10.0.0.2", 5),
                            CAddr("::ffff:10:0:0:2", 7),
                        ],
                        2,
                    ),
                ],
                [],
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=5,
                    min_count=4,
                    max_count=4,
                    node_name="node3",
                    node_index=1,
                ),
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address="::ffff:10:0:0:3",
                    expected_address_type="IPv4",
                    link_number=5,
                ),
                fixture.error(
                    report_codes.COROSYNC_ADDRESS_IP_VERSION_WRONG_FOR_LINK,
                    address="10.0.0.3",
                    expected_address_type="IPv6",
                    link_number=7,
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=["addrX1"],
                ),
                fixture.error(
                    report_codes.COROSYNC_IP_VERSION_MISMATCH_IN_LINKS,
                    link_numbers=[3],
                ),
            ],
        )

    def test_forbidden_characters(self):
        assert_report_item_list_equal(
            config_validators.add_nodes(
                [
                    {"name": "node{3}", "addrs": ["\raddr03\n"]},
                ],
                self.fixture_coronodes_1_link,
                [],
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="node{3}",
                    option_name="node 1 name",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="\raddr03\n",
                    option_name="node address",
                    **forbidden_characters_kwargs,
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE,
                    address_list=["\raddr03\n"],
                ),
            ],
        )


class RemoveNodes(TestCase):
    fixture_nodes = [
        CNode("node1", [CAddr("addr01", 1)], 1),
        CNode("node2", [CAddr("addr02", 1)], 2),
        CNode("node3", [CAddr("addr03", 1)], 3),
        CNode("node4", [CAddr("addr04", 1)], 4),
    ]

    def test_nonexisting_nodes(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node3", "nodeX", "nodeY", "node4"],
                self.fixture_nodes,
                (None, None, None, None),
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND, node=node, searched_types=[]
                )
                for node in ["nodeX", "nodeY"]
            ],
        )

    def test_all_nodes(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node3", "node1", "node2", "node4"],
                self.fixture_nodes,
                (None, None, None, None),
            ),
            [fixture.error(report_codes.CANNOT_REMOVE_ALL_CLUSTER_NODES)],
        )

    def test_qdevice_tie_breaker_none(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node4"], self.fixture_nodes, ("net", {}, None, None)
            ),
            [],
        )

    def test_qdevice_tie_breaker_generic(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node4"],
                self.fixture_nodes,
                ("net", {"tie_breaker": "highest"}, None, None),
            ),
            [],
        )

    def test_qdevice_tie_breaker_kept(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node4"],
                self.fixture_nodes,
                ("net", {"tie_breaker": "3"}, None, None),
            ),
            [],
        )

    def test_qdevice_tie_breaker_removed(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node4"],
                self.fixture_nodes,
                ("net", {"tie_breaker": "4"}, None, None),
            ),
            [
                fixture.error(
                    report_codes.NODE_USED_AS_TIE_BREAKER,
                    node="node4",
                    node_id=4,
                ),
            ],
        )

    def test_more_errors(self):
        assert_report_item_list_equal(
            config_validators.remove_nodes(
                ["node3", "node1", "node2", "node4", "nodeX"],
                self.fixture_nodes,
                ("net", {"tie_breaker": "4"}, None, None),
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND, node="nodeX", searched_types=[]
                ),
                fixture.error(report_codes.CANNOT_REMOVE_ALL_CLUSTER_NODES),
                fixture.error(
                    report_codes.NODE_USED_AS_TIE_BREAKER,
                    node="node4",
                    node_id=4,
                ),
            ],
        )



import re

from . import const
from .error import InvalidResourceAgentName
from .types import (
    ResourceAgentMetadata,
    ResourceAgentName,
)


def split_resource_agent_name(full_agent_name: str) -> ResourceAgentName:
    """
    Parse a resource agent name string to a structure

    full_agent_name -- agent name to parse
    """
    # Full_agent_name could be for example "systemd:lvm2-pvscan@252:2".
    # Note the second colon is not a separator between a provider and a type.
    # This regexp only matches systemd and service agents with instances.
    # Systemd and service agents without instances are matched by a generic
    # regexp below.
    match = re.match(
        "^(?P<standard>systemd|service):(?P<type>[^:@]+@.*)$",
        full_agent_name,
    )
    if match:
        return ResourceAgentName(
            match.group("standard"), None, match.group("type")
        )

    # This allows even bad combinations like "systemd:provider:agent" or
    # "ocf:agent". Those are caught by checks bellow to keep the regexp simple.
    match = re.match(
        "^(?P<standard>[^:]+)(:(?P<provider>[^:]+))?:(?P<type>[^:]+)$",
        full_agent_name,
    )
    if not match:
        raise InvalidResourceAgentName(full_agent_name)

    standard = match.group("standard")
    provider = match.group("provider") if match.group("provider") else None
    agent_type = match.group("type")

    # These are all the standards valid in a CIB. To get a list of standards
    # supported by pacemaker in the local environment, use result of
    # "crm_resource --list-standards" or "list_resource_agents_standards"
    allowed_standards = {
        "heartbeat",
        "lsb",
        "nagios",
        "ocf",
        "service",
        "stonith",
        "systemd",
        "upstart",
    }
    if standard not in allowed_standards:
        raise InvalidResourceAgentName(full_agent_name)
    if standard == "ocf" and not provider:
        raise InvalidResourceAgentName(full_agent_name)
    if standard != "ocf" and provider:
        raise InvalidResourceAgentName(full_agent_name)

    return ResourceAgentName(standard, provider, agent_type)


def name_to_void_metadata(name: ResourceAgentName) -> ResourceAgentMetadata:
    """
    Create a non-existent agent metadata for a given agent name

    name -- agent name to put into agent metadata
    """
    return ResourceAgentMetadata(
        name,
        agent_exists=False,
        ocf_version=const.OCF_1_0,
        shortdesc=None,
        longdesc=None,
        parameters=[],
        actions=[],
    )



from collections import defaultdict
from shlex import quote
from typing import (
    Container,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import (
    INDENT_STEP,
    format_wrap_for_terminal,
)
from pcs.cli.nvset import nvset_dto_to_lines
from pcs.cli.reports.output import warn
from pcs.cli.resource_agent import (
    get_resource_agent_full_name,
    is_stonith,
)
from pcs.common import resource_agent
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.pacemaker.resource.bundle import (
    CibResourceBundleContainerRuntimeOptionsDto,
    CibResourceBundleDto,
    CibResourceBundleNetworkOptionsDto,
    CibResourceBundlePortMappingDto,
    CibResourceBundleStorageMappingDto,
)
from pcs.common.pacemaker.resource.clone import CibResourceCloneDto
from pcs.common.pacemaker.resource.group import CibResourceGroupDto
from pcs.common.pacemaker.resource.list import ListCibResourcesDto
from pcs.common.pacemaker.resource.operations import (
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
    CibResourceOperationDto,
)
from pcs.common.pacemaker.resource.primitive import CibResourcePrimitiveDto
from pcs.common.resource_agent.dto import ResourceAgentNameDto
from pcs.common.str_tools import (
    format_name_value_list,
    format_optional,
    indent,
)


def _get_ocf_check_level_from_operation(
    operation_dto: CibResourceOperationDto,
) -> Optional[str]:
    for nvset in operation_dto.instance_attributes:
        for nvpair in nvset.nvpairs:
            if nvpair.name == OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME:
                return nvpair.value
    return None


def _resource_operation_to_pairs(
    operation_dto: CibResourceOperationDto,
) -> List[Tuple[str, str]]:
    pairs = [("interval", operation_dto.interval)]
    if operation_dto.id:
        pairs.append(("id", operation_dto.id))
    if operation_dto.start_delay:
        pairs.append(("start-delay", operation_dto.start_delay))
    elif operation_dto.interval_origin:
        pairs.append(("interval-origin", operation_dto.interval_origin))
    if operation_dto.timeout:
        pairs.append(("timeout", operation_dto.timeout))
    if operation_dto.enabled:
        pairs.append(("enabled", _bool_to_cli_value(operation_dto.enabled)))
    if operation_dto.record_pending:
        pairs.append(
            ("record-pending", _bool_to_cli_value(operation_dto.record_pending))
        )
    if operation_dto.role:
        pairs.append(("role", operation_dto.role))
    if operation_dto.on_fail:
        pairs.append(("on-fail", operation_dto.on_fail))
    ocf_check_level = _get_ocf_check_level_from_operation(operation_dto)
    if ocf_check_level:
        pairs.append((OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME, ocf_check_level))
    return pairs


def _resource_operation_to_str(
    operation_dto: CibResourceOperationDto,
) -> List[str]:
    lines = format_name_value_list(
        [
            pair
            for pair in _resource_operation_to_pairs(operation_dto)
            if pair[0] != "id"
        ]
    )
    # TODO: add support for meta and instance attributes once it is supported
    # by pcs
    return [
        "{name}:{id}".format(
            name=operation_dto.name,
            id=format_optional(operation_dto.id, " {}"),
        )
    ] + indent(lines, indent_step=INDENT_STEP)


def resource_agent_metadata_to_text(
    metadata: resource_agent.dto.ResourceAgentMetadataDto,
    default_operations: List[CibResourceOperationDto],
    verbose: bool = False,
) -> List[str]:
    # pylint: disable=too-many-branches
    output = []
    _is_stonith = is_stonith(metadata.name)
    agent_name = (
        metadata.name.type
        if _is_stonith
        else get_resource_agent_full_name(metadata.name)
    )
    if metadata.shortdesc:
        output.extend(
            format_wrap_for_terminal(
                "{agent_name} - {shortdesc}".format(
                    agent_name=agent_name,
                    shortdesc=metadata.shortdesc.replace("\n", " "),
                ),
            )
        )
    else:
        output.append(agent_name)

    if metadata.longdesc:
        output.append("")
        output.extend(
            format_wrap_for_terminal(
                metadata.longdesc.replace("\n", " "), subsequent_indent=0
            )
        )

    params = []
    for param in metadata.parameters:
        if not verbose and (param.advanced or param.deprecated):
            continue
        param_title = [param.name]
        if param.deprecated_by:
            param_title.append(
                "(deprecated by {})".format(", ".join(param.deprecated_by))
            )
        elif param.deprecated:
            param_title.append("(deprecated)")
        if param.required:
            param_title.append("(required)")
        if param.unique_group:
            if param.unique_group.startswith(
                resource_agent.const.DEFAULT_UNIQUE_GROUP_PREFIX
            ):
                param_title.append("(unique)")
            else:
                param_title.append(
                    "(unique group: {})".format(param.unique_group)
                )
        desc = ""
        if param.longdesc:
            desc = param.longdesc.replace("\n", " ")
        if not desc and param.shortdesc:
            desc = param.shortdesc.replace("\n", " ")
        if not desc:
            desc = "No description available"
        if param.deprecated_desc:
            desc += "DEPRECATED: {param.deprecated_desc}"
        params.append("{}: {}".format(" ".join(param_title), desc))
    if params:
        output.append("")
        if _is_stonith:
            output.append("Stonith options:")
        else:
            output.append("Resource options:")
        output.extend(indent(params, indent_step=INDENT_STEP))

    operations = []
    for operation in default_operations:
        operations.extend(_resource_operation_to_str(operation))

    if operations:
        output.append("")
        output.append("Default operations:")
        output.extend(indent(operations, indent_step=INDENT_STEP))

    return output


class ResourcesConfigurationFacade:
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        primitives: Sequence[CibResourcePrimitiveDto],
        groups: Sequence[CibResourceGroupDto],
        clones: Sequence[CibResourceCloneDto],
        bundles: Sequence[CibResourceBundleDto],
        filtered_ids: Iterable[str] = (),
        only_stonith: Optional[bool] = None,
    ) -> None:
        self._primitives = primitives
        self._groups = groups
        self._clones = clones
        self._bundles = bundles
        self._filtered_ids = frozenset(filtered_ids)
        self._only_stonith = only_stonith
        self._primitives_map = {res.id: res for res in self._primitives}
        self._bundles_map = {res.id: res for res in self._bundles}
        self._clones_map = {res.id: res for res in self._clones}
        self._groups_map = {res.id: res for res in self._groups}
        self._child_parent_map: Dict[str, str] = {}
        self._parent_child_map: Dict[str, List[str]] = defaultdict(list)
        for bundle_dto in self._bundles:
            if bundle_dto.member_id:
                self._set_parent(bundle_dto.member_id, bundle_dto.id)
        for clone_dto in self._clones:
            self._set_parent(clone_dto.member_id, clone_dto.id)
        for group_dto in self._groups:
            for primitive_id in group_dto.member_ids:
                self._set_parent(primitive_id, group_dto.id)

    @classmethod
    def from_resources_dto(
        cls, resources_dto: ListCibResourcesDto
    ) -> "ResourcesConfigurationFacade":
        return cls(
            resources_dto.primitives,
            resources_dto.groups,
            resources_dto.clones,
            resources_dto.bundles,
        )

    def get_parent_id(self, res_id: str) -> Optional[str]:
        return self._child_parent_map.get(res_id)

    def _get_children_ids(self, res_id: str) -> List[str]:
        return self._parent_child_map[res_id]

    def _set_parent(self, child_id: str, parent_id: str) -> None:
        if self.get_parent_id(child_id):
            raise AssertionError("invalid data")
        self._child_parent_map[child_id] = parent_id
        self._parent_child_map[parent_id].append(child_id)

    def get_primitive_dto(
        self, obj_id: str
    ) -> Optional[CibResourcePrimitiveDto]:
        return self._primitives_map.get(obj_id)

    def get_group_dto(self, obj_id: str) -> Optional[CibResourceGroupDto]:
        return self._groups_map.get(obj_id)

    def _get_any_resource_dto(
        self, obj_id: str
    ) -> Optional[
        Union[
            CibResourcePrimitiveDto,
            CibResourceGroupDto,
            CibResourceCloneDto,
            CibResourceBundleDto,
        ]
    ]:
        return (
            self._primitives_map.get(obj_id)
            or self._bundles_map.get(obj_id)
            or self._clones_map.get(obj_id)
            or self._groups_map.get(obj_id)
        )

    @property
    def filtered_ids(self) -> Container[str]:
        return self._filtered_ids

    @property
    def primitives(self) -> Sequence[CibResourcePrimitiveDto]:
        return self._primitives

    @property
    def clones(self) -> Sequence[CibResourceCloneDto]:
        return self._clones

    @property
    def groups(self) -> Sequence[CibResourceGroupDto]:
        return self._groups

    @property
    def bundles(self) -> Sequence[CibResourceBundleDto]:
        return self._bundles

    def filter_stonith(
        self, allow_stonith: bool
    ) -> "ResourcesConfigurationFacade":
        primitives = [
            primitive
            for primitive in self._primitives
            if is_stonith(primitive.agent_name) == allow_stonith
            or (not allow_stonith and self.get_parent_id(primitive.id))
        ]
        if allow_stonith:
            return self.__class__(
                primitives=primitives,
                clones=[],
                bundles=[],
                groups=[],
                filtered_ids=self._filtered_ids,
                only_stonith=True,
            )
        return self.__class__(
            primitives=primitives,
            clones=self._clones,
            bundles=self._bundles,
            groups=self._groups,
            filtered_ids=self._filtered_ids,
            only_stonith=False,
        )

    def filter_resources(
        self,
        resource_ids_to_find: Iterable[str],
        allow_warnings: bool,
    ) -> "ResourcesConfigurationFacade":
        # pylint: disable=too-many-branches
        if not resource_ids_to_find:
            return self
        primitives = set()
        groups = set()
        clones = set()
        bundles = set()

        label = "resource/stonith device"
        if self._only_stonith is True:
            label = "stonith device"
        elif self._only_stonith is False:
            label = "resource"

        ids_to_process = set(resource_ids_to_find)
        processed_ids = set()

        while ids_to_process:
            resource_id = ids_to_process.pop()
            if resource_id in processed_ids:
                continue
            resource_dto = self._get_any_resource_dto(resource_id)
            if resource_dto is None:
                if allow_warnings:
                    warn(f"Unable to find {label} '{resource_id}'")
                continue
            processed_ids.add(resource_id)
            ids_to_process.update(self._get_children_ids(resource_id))
            if isinstance(resource_dto, CibResourcePrimitiveDto):
                primitives.add(resource_id)
            elif isinstance(resource_dto, CibResourceGroupDto):
                groups.add(resource_id)
            elif isinstance(resource_dto, CibResourceCloneDto):
                clones.add(resource_id)
            elif isinstance(resource_dto, CibResourceBundleDto):
                bundles.add(resource_id)
            else:
                AssertionError()

        if not processed_ids:
            raise CmdLineInputError(f"No {label} found")
        return self.__class__(
            primitives=sorted(
                (self._primitives_map[res_id] for res_id in primitives),
                key=lambda obj: obj.id,
            ),
            clones=sorted(
                (self._clones_map[res_id] for res_id in clones),
                key=lambda obj: obj.id,
            ),
            groups=sorted(
                (self._groups_map[res_id] for res_id in groups),
                key=lambda obj: obj.id,
            ),
            bundles=sorted(
                (self._bundles_map[res_id] for res_id in bundles),
                key=lambda obj: obj.id,
            ),
            filtered_ids=set(resource_ids_to_find) & set(processed_ids),
            only_stonith=self._only_stonith,
        )


def _resource_agent_name_to_text(
    resource_agent_name_dto: ResourceAgentNameDto,
) -> str:
    output = f"class={resource_agent_name_dto.standard}"
    if resource_agent_name_dto.provider:
        output += f" provider={resource_agent_name_dto.provider}"
    output += f" type={resource_agent_name_dto.type}"
    return output


def _nvset_to_text(label: str, nvsets: Sequence[CibNvsetDto]) -> List[str]:
    if nvsets and nvsets[0].nvpairs:
        return nvset_dto_to_lines(nvset=nvsets[0], nvset_label=label)
    return []


def _resource_description_to_text(desc: Optional[str]) -> List[str]:
    if desc:
        return [f"Description: {desc}"]
    return []


def _resource_primitive_to_text(
    primitive_dto: CibResourcePrimitiveDto,
) -> List[str]:
    output = (
        _resource_description_to_text(primitive_dto.description)
        + _nvset_to_text("Attributes", primitive_dto.instance_attributes)
        + _nvset_to_text("Meta Attributes", primitive_dto.meta_attributes)
        + _nvset_to_text("Utilization", primitive_dto.utilization)
    )
    if primitive_dto.operations:
        operation_lines: List[str] = []
        for operation_dto in primitive_dto.operations:
            operation_lines.extend(_resource_operation_to_str(operation_dto))
        output.extend(
            ["Operations:"] + indent(operation_lines, indent_step=INDENT_STEP)
        )

    return [
        "Resource: {res_id} ({res_type})".format(
            res_id=primitive_dto.id,
            res_type=_resource_agent_name_to_text(primitive_dto.agent_name),
        )
    ] + indent(output, indent_step=INDENT_STEP)


def _resource_group_to_text(
    group_dto: CibResourceGroupDto,
    resources_facade: ResourcesConfigurationFacade,
) -> List[str]:
    output = (
        _resource_description_to_text(group_dto.description)
        + _nvset_to_text("Attributes", group_dto.instance_attributes)
        + _nvset_to_text("Meta Attributes", group_dto.meta_attributes)
    )
    for primitive_id in group_dto.member_ids:
        primitive_dto = resources_facade.get_primitive_dto(primitive_id)
        if primitive_dto is None:
            raise CmdLineInputError(
                f"Invalid data: group {group_dto.id} has no children"
            )
        output.extend(_resource_primitive_to_text(primitive_dto))
    return [f"Group: {group_dto.id}"] + indent(output, indent_step=INDENT_STEP)


def _resource_clone_to_text(
    clone_dto: CibResourceCloneDto,
    resources_facade: ResourcesConfigurationFacade,
) -> List[str]:
    output = (
        _resource_description_to_text(clone_dto.description)
        + _nvset_to_text("Attributes", clone_dto.instance_attributes)
        + _nvset_to_text("Meta Attributes", clone_dto.meta_attributes)
    )
    primitive_dto = resources_facade.get_primitive_dto(clone_dto.member_id)
    group_dto = resources_facade.get_group_dto(clone_dto.member_id)
    if primitive_dto is not None:
        output.extend(_resource_primitive_to_text(primitive_dto))
    elif group_dto is not None:
        output.extend(_resource_group_to_text(group_dto, resources_facade))
    else:
        raise CmdLineInputError(
            f"Invalid data: clone {clone_dto.id} has no children"
        )
    return [f"Clone: {clone_dto.id}"] + indent(output, indent_step=INDENT_STEP)


def _resource_bundle_container_options_to_pairs(
    options: CibResourceBundleContainerRuntimeOptionsDto,
) -> List[Tuple[str, str]]:
    option_list = [("image", options.image)]
    if options.replicas:
        option_list.append(("replicas", str(options.replicas)))
    if options.replicas_per_host:
        option_list.append(
            ("replicas-per-host", str(options.replicas_per_host))
        )
    if options.promoted_max:
        option_list.append(("promoted-max", str(options.promoted_max)))
    if options.run_command:
        option_list.append(("run-command", options.run_command))
    if options.network:
        option_list.append(("network", options.network))
    if options.options:
        option_list.append(("options", options.options))
    return option_list


def _bool_to_cli_value(value: bool) -> str:
    return "1" if value else "0"


def _resource_bundle_network_options_to_pairs(
    bundle_network_dto: Optional[CibResourceBundleNetworkOptionsDto],
) -> List[Tuple[str, str]]:
    network_options: List[Tuple[str, str]] = []
    if not bundle_network_dto:
        return network_options
    if bundle_network_dto.ip_range_start:
        network_options.append(
            ("ip-range-start", bundle_network_dto.ip_range_start)
        )
    if bundle_network_dto.control_port:
        network_options.append(
            ("control-port", str(bundle_network_dto.control_port))
        )
    if bundle_network_dto.host_interface:
        network_options.append(
            ("host-interface", bundle_network_dto.host_interface)
        )
    if bundle_network_dto.host_netmask:
        network_options.append(
            ("host-netmask", str(bundle_network_dto.host_netmask))
        )
    if bundle_network_dto.add_host is not None:
        network_options.append(
            ("add-host", _bool_to_cli_value(bundle_network_dto.add_host))
        )
    return network_options


def _resource_bundle_port_mapping_to_pairs(
    bundle_net_port_mapping_dto: CibResourceBundlePortMappingDto,
) -> List[Tuple[str, str]]:
    mapping = []
    if bundle_net_port_mapping_dto.port:
        mapping.append(("port", str(bundle_net_port_mapping_dto.port)))
    if bundle_net_port_mapping_dto.internal_port:
        mapping.append(
            ("internal-port", str(bundle_net_port_mapping_dto.internal_port))
        )
    if bundle_net_port_mapping_dto.range:
        mapping.append(("range", bundle_net_port_mapping_dto.range))
    return mapping


def _resource_bundle_network_port_mapping_to_str(
    bundle_net_port_mapping_dto: CibResourceBundlePortMappingDto,
) -> List[str]:
    output = format_name_value_list(
        _resource_bundle_port_mapping_to_pairs(bundle_net_port_mapping_dto)
    )
    if output and bundle_net_port_mapping_dto.id:
        output.append(f"({bundle_net_port_mapping_dto.id})")
    return output


def _resource_bundle_network_to_text(
    bundle_network_dto: Optional[CibResourceBundleNetworkOptionsDto],
) -> List[str]:
    network_options = _resource_bundle_network_options_to_pairs(
        bundle_network_dto
    )
    if network_options:
        return [
            " ".join(["Network:"] + format_name_value_list(network_options))
        ]
    return []


def _resource_bundle_port_mappings_to_text(
    bundle_port_mappings: Sequence[CibResourceBundlePortMappingDto],
) -> List[str]:
    port_mappings = [
        " ".join(_resource_bundle_network_port_mapping_to_str(port_mapping_dto))
        for port_mapping_dto in bundle_port_mappings
    ]
    if port_mappings:
        return ["Port Mapping:"] + indent(
            port_mappings, indent_step=INDENT_STEP
        )
    return []


def _resource_bundle_storage_mapping_to_pairs(
    storage_mapping: CibResourceBundleStorageMappingDto,
) -> List[Tuple[str, str]]:
    mapping = []
    if storage_mapping.source_dir:
        mapping.append(("source-dir", storage_mapping.source_dir))
    if storage_mapping.source_dir_root:
        mapping.append(("source-dir-root", storage_mapping.source_dir_root))
    mapping.append(("target-dir", storage_mapping.target_dir))
    if storage_mapping.options:
        mapping.append(("options", storage_mapping.options))
    return mapping


def _resource_bundle_storage_mapping_to_str(
    storage_mapping: CibResourceBundleStorageMappingDto,
) -> List[str]:
    return format_name_value_list(
        _resource_bundle_storage_mapping_to_pairs(storage_mapping)
    ) + [f"({storage_mapping.id})"]


def _resource_bundle_storage_to_text(
    storage_mappings: Sequence[CibResourceBundleStorageMappingDto],
) -> List[str]:
    if not storage_mappings:
        return []
    output = []
    for storage_mapping in storage_mappings:
        output.append(
            " ".join(_resource_bundle_storage_mapping_to_str(storage_mapping))
        )
    return ["Storage Mapping:"] + indent(output, indent_step=INDENT_STEP)


def _resource_bundle_to_text(
    bundle_dto: CibResourceBundleDto,
    resources_facade: ResourcesConfigurationFacade,
) -> List[str]:
    container_options = []
    if bundle_dto.container_type and bundle_dto.container_options:
        container_options.append(
            " ".join(
                ["{}:".format(str(bundle_dto.container_type).capitalize())]
                + format_name_value_list(
                    _resource_bundle_container_options_to_pairs(
                        bundle_dto.container_options
                    )
                )
            )
        )
    output = (
        _resource_description_to_text(bundle_dto.description)
        + container_options
        + _resource_bundle_network_to_text(bundle_dto.network)
        + _resource_bundle_port_mappings_to_text(bundle_dto.port_mappings)
        + _resource_bundle_storage_to_text(bundle_dto.storage_mappings)
        + _nvset_to_text("Meta Attributes", bundle_dto.meta_attributes)
    )
    if bundle_dto.member_id:
        primitive_dto = resources_facade.get_primitive_dto(bundle_dto.member_id)
        if primitive_dto is None:
            raise CmdLineInputError(
                f"Invalid data: bundle '{bundle_dto.id}' has inner primitive "
                f"resource with id '{bundle_dto.member_id}' which was not found"
            )
        output.extend(_resource_primitive_to_text(primitive_dto))
    return [f"Bundle: {bundle_dto.id}"] + indent(
        output, indent_step=INDENT_STEP
    )


def resources_to_text(
    resources_facade: ResourcesConfigurationFacade,
) -> List[str]:
    def _is_allowed_to_display_fn(res_id: str) -> bool:
        if resources_facade.filtered_ids:
            return res_id in resources_facade.filtered_ids
        return resources_facade.get_parent_id(res_id) is None

    output = []
    for primitive_dto in resources_facade.primitives:
        if _is_allowed_to_display_fn(primitive_dto.id):
            output.extend(_resource_primitive_to_text(primitive_dto))
    for group_dto in resources_facade.groups:
        if _is_allowed_to_display_fn(group_dto.id):
            output.extend(_resource_group_to_text(group_dto, resources_facade))
    for clone_dto in resources_facade.clones:
        if _is_allowed_to_display_fn(clone_dto.id):
            output.extend(_resource_clone_to_text(clone_dto, resources_facade))
    for bundle_dto in resources_facade.bundles:
        if _is_allowed_to_display_fn(bundle_dto.id):
            output.extend(
                _resource_bundle_to_text(bundle_dto, resources_facade)
            )

    return output


def _nvset_to_cmd(
    label: Optional[str],
    nvsets: Sequence[CibNvsetDto],
) -> List[str]:
    if nvsets and nvsets[0].nvpairs:
        options = _pairs_to_cmd(
            (nvpair.name, nvpair.value) for nvpair in nvsets[0].nvpairs
        )
        if label:
            options = f"{label} {options}"
        return [options]
    return []


def _resource_operation_to_cmd(
    operations: Sequence[CibResourceOperationDto],
) -> List[str]:
    if not operations:
        return []
    cmd = []
    for op in operations:
        cmd.append(
            "{name} {options}".format(
                name=op.name,
                options=_pairs_to_cmd(_resource_operation_to_pairs(op)),
            )
        )
    return ["op"] + indent(cmd, indent_step=INDENT_STEP)


def _resource_primitive_to_cmd(
    primitive_dto: CibResourcePrimitiveDto,
    bundle_id: Optional[str],
) -> List[List[str]]:
    _is_stonith = is_stonith(primitive_dto.agent_name)
    options = (
        _nvset_to_cmd(None, primitive_dto.instance_attributes)
        + _resource_operation_to_cmd(primitive_dto.operations)
        + _nvset_to_cmd("meta", primitive_dto.meta_attributes)
    )
    if bundle_id:
        options.append(f"bundle {bundle_id}")

    output = [
        [
            _options_to_cmd(
                [
                    "pcs",
                    "stonith" if _is_stonith else "resource",
                    "create",
                    "--no-default-ops",
                    "--force",
                    "--",
                    primitive_dto.id,
                    (
                        primitive_dto.agent_name.type
                        if _is_stonith
                        else get_resource_agent_full_name(
                            primitive_dto.agent_name
                        )
                    ),
                ]
            )
        ]
        + indent(options, indent_step=INDENT_STEP)
    ]
    utilization_cmd_params = _nvset_to_cmd(None, primitive_dto.utilization)
    if utilization_cmd_params:
        output.append(
            [
                _options_to_cmd(
                    ["pcs", "resource", "utilization", primitive_dto.id]
                )
            ]
            + indent(utilization_cmd_params, indent_step=INDENT_STEP)
        )

    return output


def _options_to_cmd(options: Iterable[str]) -> str:
    return " ".join(quote(option) for option in options)


def _pairs_to_cmd(pairs: Iterable[Tuple[str, str]]) -> str:
    return _options_to_cmd("=".join(item) for item in pairs)


def _resource_bundle_to_cmd(
    bundle_dto: CibResourceBundleDto,
) -> List[List[str]]:
    if not (bundle_dto.container_type and bundle_dto.container_options):
        return []
    options = [
        _options_to_cmd(["container", str(bundle_dto.container_type)])
    ] + indent(
        [
            _pairs_to_cmd(
                _resource_bundle_container_options_to_pairs(
                    bundle_dto.container_options
                )
            )
        ],
        indent_step=INDENT_STEP,
    )
    network_options = _pairs_to_cmd(
        _resource_bundle_network_options_to_pairs(bundle_dto.network)
    )
    if network_options:
        options.append(f"network {network_options}")
    for port_mapping in bundle_dto.port_mappings:
        options.append(
            "port-map {}".format(
                _pairs_to_cmd(
                    _resource_bundle_port_mapping_to_pairs(port_mapping)
                )
            )
        )
    for storage_mapping in bundle_dto.storage_mappings:
        options.append(
            "storage-map {}".format(
                _pairs_to_cmd(
                    _resource_bundle_storage_mapping_to_pairs(storage_mapping)
                )
            )
        )
    options.extend(_nvset_to_cmd("meta", bundle_dto.meta_attributes))
    return [
        [
            _options_to_cmd(
                ["pcs", "resource", "bundle", "create", bundle_dto.id]
            )
        ]
        + indent(options, indent_step=INDENT_STEP)
    ]


def _resource_group_to_cmd(group_dto: CibResourceGroupDto) -> List[List[str]]:
    output = []
    output.append(
        [_options_to_cmd(["pcs", "resource", "group", "add", group_dto.id])]
        + indent(
            [_options_to_cmd(group_dto.member_ids)],
            indent_step=INDENT_STEP,
        )
    )
    meta_options = _nvset_to_cmd(None, group_dto.meta_attributes)
    if meta_options:
        output.append(
            [_options_to_cmd(["pcs", "resource", "meta", group_dto.id])]
            + indent(meta_options, indent_step=INDENT_STEP)
        )
    return output


def _resource_clone_to_cmd(clone_dto: CibResourceCloneDto) -> List[List[str]]:
    return [
        [
            _options_to_cmd(
                ["pcs", "resource", "clone", clone_dto.member_id, clone_dto.id]
            )
        ]
        + indent(
            _nvset_to_cmd("meta", clone_dto.meta_attributes),
            indent_step=INDENT_STEP,
        )
    ]


def resources_to_cmd(
    resources_facade: ResourcesConfigurationFacade,
) -> List[List[str]]:
    output: List[List[str]] = []
    primitives_created_with_bundle = set()
    for bundle_dto in resources_facade.bundles:
        if not (bundle_dto.container_type and bundle_dto.container_options):
            warn(
                f"Bundle resource '{bundle_dto.id}' uses unsupported container "
                "type, therefore pcs is unable to create it. The resource will be omitted."
            )
            continue
        output.extend(_resource_bundle_to_cmd(bundle_dto))
        if bundle_dto.member_id:
            primitive_dto = resources_facade.get_primitive_dto(
                bundle_dto.member_id
            )
            if primitive_dto is None:
                raise CmdLineInputError(
                    f"Invalid data: bundle '{bundle_dto.id}' has inner "
                    f"primitive resource with id '{bundle_dto.member_id}' "
                    "which was not found"
                )
            output.extend(
                _resource_primitive_to_cmd(primitive_dto, bundle_dto.id)
            )
            primitives_created_with_bundle.add(bundle_dto.member_id)
    for primitive_dto in resources_facade.primitives:
        if primitive_dto.id not in primitives_created_with_bundle:
            output.extend(_resource_primitive_to_cmd(primitive_dto, None))
    for group_dto in resources_facade.groups:
        output.extend(_resource_group_to_cmd(group_dto))
    for clone_dto in resources_facade.clones:
        output.extend(_resource_clone_to_cmd(clone_dto))
    return output



import os
from unittest import TestCase

from pcs.common import const
from pcs.common.str_tools import format_list

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_dir,
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class ParseArgvDashDash(TestCase, AssertPcsMixin):
    # The command will fail, that's ok. We are interested only in the argv
    # parsing.
    cmd = "constraint colocation add R1 with R2".split()

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_misc_dashdash")
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.allowed_roles = format_list(const.PCMK_ROLES)

    def tearDown(self):
        self.temp_cib.close()

    def test_negative_int(self):
        self.assert_pcs_fail(
            self.cmd + ["-123"],
            outdent(
                """\
                Warning: Using '-123' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                Error: Resource 'R1' does not exist
                """
            ),
        )

    def test_negative_float(self):
        self.assert_pcs_fail(
            self.cmd + ["-12.3"],
            outdent(
                f"""\
                Warning: Using '-12.3' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )

    def test_negative_infinity(self):
        self.assert_pcs_fail(
            self.cmd + ["-inFIniTY"],
            outdent(
                f"""\
                Warning: Using '-inFIniTY' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )

    def test_negative_int_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-123"],
            outdent(
                """\
                Error: Resource 'R1' does not exist
                """
            ),
        )

    def test_negative_float_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-12.3"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )

    def test_negative_infinity_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-inFIniTY"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )


class EmptyCibIsPcmk2Compatible(TestCase, AssertPcsMixin):
    # This test verifies that a default empty CIB created by pcs when -f points
    # to an empty file conforms to minimal schema version supported by
    # pacemaker 2.0. If pcs prints a message that CIB schema has been upgraded,
    # then the test fails and shows there is a bug. Bundle with promoted-max
    # requires CIB compliant with schema 3.1, which was introduced in pacemaker
    # 2.0.0.
    def setUp(self):
        self.cib_dir = get_tmp_dir("tier1_misc_empty_cib")
        self.pcs_runner = PcsRunner(os.path.join(self.cib_dir.name, "cib.xml"))

    def tearDown(self):
        self.cib_dir.cleanup()

    def test_success(self):
        self.assert_pcs_success(
            "resource bundle create b container docker image=my.img promoted-max=1".split(),
            "",
        )
